from typing import List, Optional

from pycubexr.classes.region import Region


class CNode(object):
    _children: List['CNode']

    def __init__(
            self,
            *,
            _id: int,
            callee_region_id: int
    ):
        self._children = []
        self.id = _id
        self.callee_region_id = callee_region_id
        self.region: Optional[Region] = None
        self.parameters = {}
        self.parent: Optional['CNode'] = None

    def get_children(self):
        return self._children

    def add_child(self, child: 'CNode'):
        self._children.append(child)

    def get_all_children(self, with_self=True) -> List['CNode']:
        cnodes = []
        if with_self:
            cnodes.append(self)
        for child in self._children:
            cnodes += child.get_all_children()
        return cnodes

    def __repr__(self):
        if self.region is None:
            return f'CNode<{self.id}, region_id:{self.callee_region_id}, children:{self._children}>'
        else:
            return f'CNode<{self.id}, region:<{self.callee_region_id}, {self.region.name}>, children:{self._children}>'



import re
import struct
import zlib
from typing import List, BinaryIO, Any

import numpy as np

from pycubexr.classes.values import convert_type
from pycubexr.utils.exceptions import UnsupportedMetricFormatError
from pycubexr.utils.metric_formats import METRIC_FORMATS

DATA_HEADER: bytes = b'CUBEX.DATA'
ZDATA_HEADER: bytes = b'ZCUBEX.DATA'


def parse_data(
        *,
        data_file: BinaryIO,
        # From the anchor.xml. NOT the Python struct data types, e.g., "i"
        data_type: str,
        # Either "<" or ">"
        endianness_format_char: str,
        allow_full_uint64_values: bool = False,
        use_numpy=False
) -> List[Any]:
    # Verify the data file header
    header = data_file.read(len(DATA_HEADER))
    if header != DATA_HEADER:
        header = header + data_file.read(len(ZDATA_HEADER) - len(DATA_HEADER))
        assert header == ZDATA_HEADER
        raw = decompress_data(data_file, endianness_format_char)
    else:
        raw = data_file.read()

    internal_data_type, parameters, binary_type = _get_metric_format(data_type)
    if internal_data_type is None:
        raise UnsupportedMetricFormatError(data_type)

    dtype = np.dtype(binary_type).newbyteorder(endianness_format_char)

    data = np.frombuffer(raw, dtype=dtype)
    return convert_type(internal_data_type, parameters, data, allow_full_uint64_values)


def _get_metric_format(data_type):
    parameters = None
    if '(' in data_type:
        # data type has parameters
        matches = re.fullmatch(r"^(.*?)\s*\(\s*(.*?)\s*\)\s*$", data_type)
        if matches is None:
            return None, None, None
        data_type = matches[1]
        if data_type not in METRIC_FORMATS:
            return None, None, None
        parameters = re.split(r"\s*,\s*", matches[2])
        try:
            metric_format = METRIC_FORMATS[data_type](*parameters)
        except ValueError:
            return None, None, None
    elif data_type not in METRIC_FORMATS:
        return None, None, None
    else:
        metric_format = METRIC_FORMATS[data_type]
    return data_type, parameters, metric_format


def decompress_data(data_file: BinaryIO, endianness_format_char: str):
    _sizeof_long_ = struct.calcsize('q')

    raw = data_file.read(_sizeof_long_)
    number_of_entries = struct.unpack(endianness_format_char + 'q', raw)[0]

    _num_infos_ = 3
    raw = data_file.read(_num_infos_ * number_of_entries * _sizeof_long_)
    raw_headers = struct.unpack(endianness_format_char + str(_num_infos_ * number_of_entries) + 'q', raw)
    entry_headers = (raw_headers[i:i + _num_infos_] for i in range(0, len(raw_headers), _num_infos_))

    decompressed = bytes()
    for uncompressed_pos, compressed_pos, compressed_size in entry_headers:
        if compressed_size == 0:
            continue

        raw = data_file.read(compressed_size)
        decompressed += zlib.decompress(raw)

    return decompressed



import tarfile
from gzip import GzipFile
from os import PathLike
from tarfile import TarFile
from typing import List, Dict, Tuple, Union
from xml.etree import ElementTree

from pycubexr.classes import Metric, MetricValues, Region, CNode, Location
from pycubexr.parsers.anchor_xml_parser import parse_anchor_xml, AnchorXMLParseResult
from pycubexr.parsers.metrics_parser import extract_metric_values
from pycubexr.utils.caching import cached_property
from pycubexr.utils.custom_tarinfo import TarInfoWithoutCheck
from pycubexr.utils.exceptions import MissingMetricError


class CubexParser(object):
    _cubex_file: TarFile
    _cubex_filename: Union[str, PathLike]
    _anchor_result: AnchorXMLParseResult
    _metric_values: Dict[Tuple[int, bool], MetricValues]

    def __init__(self, cubex_filename: Union[str, PathLike]):
        self._cubex_filename = cubex_filename
        self._metric_values = {}

    def __enter__(self):
        try:
            self._cubex_file = tarfile.open(self._cubex_filename)
        except tarfile.ReadError:
            self._cubex_file = tarfile.open(self._cubex_filename, tarinfo=TarInfoWithoutCheck)

        self._tar_file_member_list = [x.name for x in self._cubex_file.getmembers()]

        with self._cubex_file.extractfile('anchor.xml') as anchor_file:
            xml_header = anchor_file.read(5)
            anchor_file.seek(0, 0)
            if xml_header != b"<?xml":
                # if not starting with xml assume compressed
                with GzipFile(fileobj=anchor_file) as compressed_anchor:
                    anchor = ElementTree.parse(compressed_anchor)
                    self._anchor_result = parse_anchor_xml(anchor)
            else:
                anchor = ElementTree.parse(anchor_file)
                self._anchor_result = parse_anchor_xml(anchor)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._metric_values.clear()
        self._cubex_file.close()

    def get_metric_values(
            self,
            metric: Metric,
            cache: bool = True, *,
            allow_full_uint64_values: bool = False
    ) -> MetricValues:
        """
        Retrieves the values for the specified metric.
        :param metric: The specified metric
        :param cache: Enables caching of the values to accelerate future accesses.
        :param allow_full_uint64_values: Enables the usage of UINT64 values greater than 0xFFFF_FFFF_FFFF_FBFF.
            Disabled by default to match the CubeLib behavior.
        :return: The measured values for the specified metric.
        """
        if (metric.id, allow_full_uint64_values) in self._metric_values:
            return self._metric_values[metric.id, allow_full_uint64_values]

        index_file_name = '{}.index'.format(metric.id)
        data_file_name = '{}.data'.format(metric.id)

        if index_file_name not in self._tar_file_members():
            raise MissingMetricError(metric)

        with self._cubex_file.extractfile(index_file_name) as index_file, \
                self._cubex_file.extractfile(data_file_name) as data_file:
            metric_values = extract_metric_values(
                metric=metric,
                index_file=index_file,
                data_file=data_file,
                allow_full_uint64_values=allow_full_uint64_values
            )

            assert metric_values.num_locations() == self._num_locations
            if cache:
                self._metric_values[metric.id, allow_full_uint64_values] = metric_values
            return metric_values

    @cached_property
    def _num_locations(self):
        return len(self.get_locations())

    def get_metrics(self):
        return self._anchor_result.metrics

    @cached_property
    def _metrics_dict(self):
        metrics = {}

        def walk_tree(childs):
            for metric in childs:
                metrics[metric.name] = metric
                walk_tree(metric.childs)

        walk_tree(self._anchor_result.metrics)
        return metrics

    def all_metrics(self):
        return list(self._metrics_dict.values())

    def get_metric_by_name(self, metric_name: str) -> Metric:
        return self._metrics_dict[metric_name]

    def get_region(self, cnode: CNode) -> Region:
        return self._anchor_result.regions_by_id[cnode.callee_region_id]

    def get_cnode(self, cnode_id: int) -> CNode:
        return self._anchor_result.all_cnodes.get(cnode_id)

    def get_root_cnodes(self) -> List[CNode]:
        return self._anchor_result.cnodes

    def get_region_by_name(self, name: str):
        return [region for region in self._anchor_result.regions if region.name == name][0]

    def all_cnodes(self) -> List[CNode]:
        return list(self._anchor_result.all_cnodes.values())

    def get_cnodes_for_region(self, region_id: int):
        return [cnode for cnode in self.all_cnodes() if cnode.callee_region_id == region_id]

    def get_locations(self) -> List[Location]:
        return self._anchor_result.system_tree_nodes[0].all_locations()

    def get_calltree(self, indent=0, cnode: CNode = None):
        if cnode is None:
            cnode = self._anchor_result.cnodes[0]
        call_tree_string = ""
        child_string = ""
        child_string += "-" * indent + self.get_region(cnode).name
        call_tree_string += child_string
        call_tree_string += "\n"

        for child in cnode.get_children():
            tmp = self.get_calltree(indent + 1, cnode=child)
            if tmp is not None:
                call_tree_string += tmp

        return call_tree_string

    def print_calltree(self, indent=0, cnode: CNode = None):
        if cnode is None:
            cnode = self._anchor_result.cnodes[0]

        print('\t' * indent, self.get_region(cnode).name)

        for child in cnode.get_children():
            self.print_calltree(indent + 1, cnode=child)

    def _tar_file_members(self) -> List[str]:
        return self._tar_file_member_list



import unittest
from pathlib import Path
from unittest import SkipTest

from pycubexr import CubexParser

DEEP_ENUMERATION = dict(
    [(18, 18), (55, 55), (36, 36), (52, 52), (35, 35), (82, 82), (58, 58), (41, 41), (32, 32), (6, 6), (84, 84),
     (11, 11), (12, 12), (34, 34), (42, 42), (46, 46), (13, 13), (20, 20), (77, 77), (59, 59), (63, 63), (31, 31),
     (23, 23), (24, 24), (57, 57), (79, 79), (21, 21), (10, 10), (72, 72), (40, 40), (73, 73), (27, 27), (2, 2),
     (75, 75), (56, 56), (68, 68), (81, 81), (16, 16), (80, 80), (25, 25), (37, 37), (86, 86), (61, 61), (50, 50),
     (64, 64), (0, 0), (74, 74), (29, 29), (53, 53), (30, 30), (51, 51), (26, 26), (71, 71), (33, 33), (8, 8),
     (85, 85), (19, 19), (5, 5), (47, 47), (3, 3), (38, 38), (44, 44), (49, 49), (65, 65), (43, 43), (88, 88),
     (87, 87), (45, 45), (4, 4), (7, 7), (69, 69), (1, 1), (70, 70), (39, 39), (14, 14), (17, 17), (15, 15),
     (76, 76), (60, 60), (22, 22), (67, 67), (62, 62), (66, 66), (83, 83), (48, 48), (28, 28), (9, 9), (54, 54),
     (78, 78), ])

WIDE_ENUMERATION = dict(
    [(18, 32), (55, 74), (36, 43), (52, 58), (35, 49), (82, 86), (58, 76), (41, 51), (32, 41), (6, 25), (84, 82),
     (11, 6), (12, 7), (34, 42), (46, 54), (42, 61), (13, 8), (20, 34), (77, 80), (59, 67), (63, 70), (31, 40),
     (23, 44), (24, 45), (57, 66), (79, 81), (21, 27), (10, 5), (72, 13), (40, 60), (73, 14), (27, 37), (2, 21),
     (75, 16), (68, 73), (56, 75), (81, 85), (16, 26), (80, 84), (86, 18), (37, 28), (25, 36), (61, 68), (50, 57),
     (64, 71), (0, 0), (74, 15), (29, 38), (53, 29), (30, 39), (51, 64), (71, 12), (26, 46), (33, 48), (8, 3),
     (85, 88), (19, 33), (5, 24), (47, 55), (3, 22), (38, 50), (44, 62), (49, 63), (65, 78), (43, 52), (88, 20),
     (87, 19), (45, 53), (4, 23), (7, 2), (69, 30), (1, 1), (70, 11), (39, 59), (14, 9), (17, 31), (15, 10),
     (76, 17), (60, 77), (22, 35), (67, 79), (62, 69), (66, 72), (83, 87), (48, 56), (28, 47), (9, 4), (54, 65),
     (78, 83), ])


class TestEnumeration(unittest.TestCase):
    cubex: CubexParser = None

    @classmethod
    def setUpClass(cls) -> None:
        try:
            cubex_file_path = Path("../data/cube_samples/summary.cubex").resolve()
            cls.cubex = CubexParser(cubex_file_path).__enter__()
        except FileNotFoundError as err:
            raise SkipTest("Required test file was not found.") from err

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def get_enumeration_for_metric(self, name):
        metric = self.cubex.get_metric_by_name(name)
        return metric.tree_enumeration

    def test_time(self):
        enumeration = self.get_enumeration_for_metric('time')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_execution(self):
        enumeration = self.get_enumeration_for_metric('execution')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi(self):
        enumeration = self.get_enumeration_for_metric('mpi')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_synchronization(self):
        enumeration = self.get_enumeration_for_metric('mpi_synchronization')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_sync_collective(self):
        enumeration = self.get_enumeration_for_metric('mpi_sync_collective')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_rma_synchronization(self):
        enumeration = self.get_enumeration_for_metric('mpi_rma_synchronization')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_rma_sync_active(self):
        enumeration = self.get_enumeration_for_metric('mpi_rma_sync_active')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_rma_sync_passive(self):
        enumeration = self.get_enumeration_for_metric('mpi_rma_sync_passive')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_communication(self):
        enumeration = self.get_enumeration_for_metric('mpi_communication')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_point2point(self):
        enumeration = self.get_enumeration_for_metric('mpi_point2point')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_collective(self):
        enumeration = self.get_enumeration_for_metric('mpi_collective')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_rma_communication(self):
        enumeration = self.get_enumeration_for_metric('mpi_rma_communication')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_io(self):
        enumeration = self.get_enumeration_for_metric('mpi_io')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_io_collective(self):
        enumeration = self.get_enumeration_for_metric('mpi_io_collective')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_init_exit(self):
        enumeration = self.get_enumeration_for_metric('mpi_init_exit')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_time(self):
        enumeration = self.get_enumeration_for_metric('omp_time')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_flush(self):
        enumeration = self.get_enumeration_for_metric('omp_flush')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_management(self):
        enumeration = self.get_enumeration_for_metric('omp_management')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_synchronization(self):
        enumeration = self.get_enumeration_for_metric('omp_synchronization')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_barrier(self):
        enumeration = self.get_enumeration_for_metric('omp_barrier')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_ebarrier(self):
        enumeration = self.get_enumeration_for_metric('omp_ebarrier')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_ibarrier(self):
        enumeration = self.get_enumeration_for_metric('omp_ibarrier')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_critical(self):
        enumeration = self.get_enumeration_for_metric('omp_critical')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_lock_api(self):
        enumeration = self.get_enumeration_for_metric('omp_lock_api')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_ordered(self):
        enumeration = self.get_enumeration_for_metric('omp_ordered')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_time(self):
        enumeration = self.get_enumeration_for_metric('opencl_time')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_host(self):
        enumeration = self.get_enumeration_for_metric('opencl_host')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_setup(self):
        enumeration = self.get_enumeration_for_metric('opencl_setup')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_comm(self):
        enumeration = self.get_enumeration_for_metric('opencl_comm')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_sync(self):
        enumeration = self.get_enumeration_for_metric('opencl_sync')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_kernel_launches(self):
        enumeration = self.get_enumeration_for_metric('opencl_kernel_launches')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_device(self):
        enumeration = self.get_enumeration_for_metric('opencl_device')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_opencl_kernel_executions(self):
        enumeration = self.get_enumeration_for_metric('opencl_kernel_executions')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_pthread_time(self):
        enumeration = self.get_enumeration_for_metric('pthread_time')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_pthread_management(self):
        enumeration = self.get_enumeration_for_metric('pthread_management')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_pthread_synchronization(self):
        enumeration = self.get_enumeration_for_metric('pthread_synchronization')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_pthread_mutex_api(self):
        enumeration = self.get_enumeration_for_metric('pthread_mutex_api')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_pthread_cond_api(self):
        enumeration = self.get_enumeration_for_metric('pthread_cond_api')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_overhead(self):
        enumeration = self.get_enumeration_for_metric('overhead')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_idle_threads(self):
        enumeration = self.get_enumeration_for_metric('omp_idle_threads')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_omp_limited_parallelism(self):
        enumeration = self.get_enumeration_for_metric('omp_limited_parallelism')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_visits(self):
        enumeration = self.get_enumeration_for_metric('visits')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs(self):
        enumeration = self.get_enumeration_for_metric('syncs')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_rma(self):
        enumeration = self.get_enumeration_for_metric('syncs_rma')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_fence(self):
        enumeration = self.get_enumeration_for_metric('syncs_fence')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_gats(self):
        enumeration = self.get_enumeration_for_metric('syncs_gats')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_gats_access(self):
        enumeration = self.get_enumeration_for_metric('syncs_gats_access')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_gats_exposure(self):
        enumeration = self.get_enumeration_for_metric('syncs_gats_exposure')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_syncs_locks(self):
        enumeration = self.get_enumeration_for_metric('syncs_locks')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_comms(self):
        enumeration = self.get_enumeration_for_metric('comms')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_comms_rma(self):
        enumeration = self.get_enumeration_for_metric('comms_rma')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_comms_rma_puts(self):
        enumeration = self.get_enumeration_for_metric('comms_rma_puts')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_comms_rma_gets(self):
        enumeration = self.get_enumeration_for_metric('comms_rma_gets')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes(self):
        enumeration = self.get_enumeration_for_metric('bytes')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_p2p(self):
        enumeration = self.get_enumeration_for_metric('bytes_p2p')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_sent_p2p(self):
        enumeration = self.get_enumeration_for_metric('bytes_sent_p2p')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_received_p2p(self):
        enumeration = self.get_enumeration_for_metric('bytes_received_p2p')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_coll(self):
        enumeration = self.get_enumeration_for_metric('bytes_coll')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_sent_coll(self):
        enumeration = self.get_enumeration_for_metric('bytes_sent_coll')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_received_coll(self):
        enumeration = self.get_enumeration_for_metric('bytes_received_coll')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_rma(self):
        enumeration = self.get_enumeration_for_metric('bytes_rma')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_put(self):
        enumeration = self.get_enumeration_for_metric('bytes_put')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_bytes_get(self):
        enumeration = self.get_enumeration_for_metric('bytes_get')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_ops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_ops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_iops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_iops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_irops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_irops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_iwops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_iwops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_cops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_cops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_crops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_crops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_mpi_file_cwops(self):
        enumeration = self.get_enumeration_for_metric('mpi_file_cwops')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance(self):
        enumeration = self.get_enumeration_for_metric('imbalance')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance_above(self):
        enumeration = self.get_enumeration_for_metric('imbalance_above')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance_above_single(self):
        enumeration = self.get_enumeration_for_metric('imbalance_above_single')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance_below(self):
        enumeration = self.get_enumeration_for_metric('imbalance_below')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance_below_bypass(self):
        enumeration = self.get_enumeration_for_metric('imbalance_below_bypass')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_imbalance_below_singularity(self):
        enumeration = self.get_enumeration_for_metric('imbalance_below_singularity')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_min_time(self):
        enumeration = self.get_enumeration_for_metric('min_time')
        self.assertDictEqual(DEEP_ENUMERATION, enumeration)

    def test_max_time(self):
        enumeration = self.get_enumeration_for_metric('max_time')
        self.assertDictEqual(DEEP_ENUMERATION, enumeration)

    def test_task_migration_loss(self):
        enumeration = self.get_enumeration_for_metric('task_migration_loss')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)

    def test_task_migration_win(self):
        enumeration = self.get_enumeration_for_metric('task_migration_win')
        self.assertDictEqual(WIDE_ENUMERATION, enumeration)


if __name__ == '__main__':
    unittest.main()



from typing import List

from pycubexr.classes import Location


class LocationGroup(object):
    _location_groups: List['LocationGroup']
    _locations: List[Location]

    def __init__(self, *, _id: int, name: str, rank: int, _type: str):
        self._location_groups = []
        self._locations = []
        self.id = _id
        self.name = name
        self.rank = rank
        self.type = _type

    def all_locations(self):
        locations = [x for x in self._locations]
        for location_group in self._location_groups:
            locations += location_group.all_locations()
        return locations

    def add_location_group(self, child: 'LocationGroup'):
        self._location_groups.append(child)

    def add_location(self, child: Location):
        self._locations.append(child)

    def __repr__(self):
        return 'LocationGroup<{}>'.format(self.__dict__)



import unittest
from collections import defaultdict

import numpy as np

from pycubexr.utils.exceptions import MissingMetricError

from pycubexr import CubexParser


class TestExtraPUseCase(unittest.TestCase):

    def test_weak_exclusive(self):
        self.execute()

    def test_strong_exclusive(self):
        self.execute(scaling_type='strong')

    def execute(self, scaling_type='weak', use_inclusive_measurements=False):
        aggregated_values = defaultdict(list)
        for i in range(5):
            with CubexParser('../data/miniFE/profile.cubex') as parsed:
                # iterate over all metrics
                for cube_metric in parsed.get_metrics():
                    try:
                        metric_values = parsed.get_metric_values(metric=cube_metric, cache=False)
                        # create the metrics
                        metric = cube_metric.name

                        for cnode_id in metric_values.cnode_indices:
                            cnode = parsed.get_cnode(cnode_id)
                            callpath = cnode.region.name
                            # NOTE: here we can use clustering algorithm to select only certain node level values
                            # create the measurements
                            cnode_values = metric_values.cnode_values(cnode,
                                                                      convert_to_exclusive=not use_inclusive_measurements,
                                                                      convert_to_inclusive=use_inclusive_measurements)

                            # in case of weak scaling calculate mean and median over all mpi process values
                            if scaling_type == "weak":
                                # do NOT use generator it is slower
                                if isinstance(cnode_values, np.ndarray):
                                    aggregated_values[(callpath, metric)].append(cnode_values)
                                else:
                                    try:
                                        aggregated_values[(callpath, metric)].append(cnode_values.astype(float))
                                    except AttributeError as e:
                                        print(e)
                                        print(cnode_values)
                                # else:
                                #     aggregated_values[(callpath, metric)].extend(map(float, cnode_values))

                                # in case of strong scaling calculate the sum over all mpi process values
                            elif scaling_type == "strong":
                                aggregated_values[(callpath, metric)].append(cnode_values.sum().astype(float))

                    # Take care of missing metrics
                    except MissingMetricError as e:  # @UnusedVariable
                        pass


if __name__ == '__main__':
    unittest.main()



import csv
import io
import unittest
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from zipfile import ZipFile

from pycubexr import CubexParser
from pycubexr.classes.values import CubeValues
from pycubexr.utils.exceptions import MissingMetricError


class TestMetricValues(unittest.TestCase):
    def test_time_example(self):
        cubex_file_path = Path("../data/time.p4.n2000.x1.r0/profile.cubex").resolve()
        excl_file_path = Path("../data/time.p4.n2000.x1.r0/excl.csv").resolve()
        incl_file_path = Path("../data/time.p4.n2000.x1.r0/incl.csv").resolve()
        with CubexParser(cubex_file_path) as cubex:
            with open(excl_file_path) as excl:
                self.check_against_csv_dump(cubex, excl)
            with open(incl_file_path) as incl:
                self.check_against_csv_dump(cubex, incl, inclusive=True)

    def test_call_tree_example(self):
        cubex_file_path = Path("../data/call_tree_test/call_tree_test.cubex").resolve()
        excl_file_path = Path("../data/call_tree_test/excl.csv").resolve()
        incl_file_path = Path("../data/call_tree_test/incl.csv").resolve()
        with CubexParser(cubex_file_path) as cubex:
            with open(excl_file_path) as excl:
                self.check_against_csv_dump(cubex, excl)
            with open(incl_file_path) as incl:
                self.check_against_csv_dump(cubex, incl, inclusive=True)

    def test_blast_example(self):
        cubex_file_path = Path("../data/blast.p64.r1/profile.cubex").resolve()
        excl_file_path = Path("../data/blast.p64.r1/excl.csv").resolve()
        incl_file_path = Path("../data/blast.p64.r1/incl.csv").resolve()
        with CubexParser(cubex_file_path) as cubex:
            with open(excl_file_path) as excl:
                self.check_against_csv_dump(cubex, excl)
            with open(incl_file_path) as incl:
                self.check_against_csv_dump(cubex, incl, inclusive=True)

    def test_miniFE_example(self):
        cubex_file_path = Path("../data/miniFE/profile.cubex").resolve()
        csv_file_path = Path("../data/miniFE/csv.zip").resolve()
        with CubexParser(cubex_file_path) as cubex, ZipFile(csv_file_path) as csv_zip:
            with csv_zip.open("excl.csv") as excl:
                self.check_against_csv_dump(cubex, io.TextIOWrapper(excl, encoding='utf-8', newline=''))
            with csv_zip.open("incl.csv") as incl:
                self.check_against_csv_dump(cubex, io.TextIOWrapper(incl, encoding='utf-8', newline=''), inclusive=True)

    def check_against_csv_dump(self, cubex, csv_dump, inclusive=False):
        csvreader = csv.reader(csv_dump)
        metrics = next(csvreader)  # read header
        for metric_index, metric_name in enumerate(metrics):
            if metric_index < 2:
                continue  # skip first two columns
            with self.subTest(metric=metric_name, inclusive=inclusive):
                metric = cubex.get_metric_by_name(metric_name)
                self.assertNotEqual(metric, None)
                csv_dump.seek(0)
                next(csvreader)  # skip header
                try:
                    metric_values = cubex.get_metric_values(metric, allow_full_uint64_values=True)
                    for cnode_id, rows in groupby(csvreader, itemgetter(0)):
                        cnode = cubex.get_cnode(int(cnode_id))
                        self.assertNotEqual(cnode, None)

                        values = metric_values.cnode_values(cnode, convert_to_exclusive=not inclusive,
                                                            convert_to_inclusive=inclusive)

                        original_values = [row[metric_index] for row in rows]
                        self.assertEqual(len(values), len(original_values),
                                         msg=f"at {metric_name} for cnode {cnode_id}")
                        for original_value, value in zip(original_values, values):
                            try:
                                if isinstance(value, CubeValues):
                                    value = value.astype(float)
                                pos_dot = original_value.rfind('.')
                                pos_e = original_value.find('e')
                                if pos_e >= 0:
                                    places = pos_e - pos_dot - 1
                                else:
                                    places = len(original_value) - 1 - pos_dot - 1
                                self.assertAlmostEqual(float(original_value), value, places=places,
                                                       msg=f"at {metric_name} for cnode {cnode_id} {cnode.region.name}: expected {original_value} got {value}")
                            except AssertionError:
                                raise
                except MissingMetricError:
                    self.assertTrue(all(int(r[metric_index]) == 0 for r in csvreader),
                                    msg=f"Metric {metric_name} was signaled as missing even though it should contain "
                                        f"values.")



class Location(object):
    def __init__(
            self,
            *,
            _id: int,
            name: str,
            rank: str,
            _type: str
    ):
        self.id = _id
        self.name = name
        self.rank = rank
        self.type = _type

    def __repr__(self):
        return 'Location<{}>'.format(self.__dict__)



import unittest
from pathlib import Path
from unittest import SkipTest

from pycubexr import CubexParser


class TestCube4_8WithBug(unittest.TestCase):

    def test_load(self):
        try:
            cubex_file_path = Path("../data/cube_samples/cube4.8_with_checksum_bug.cubex").resolve()
            with self.assertWarnsRegex(UserWarning, "checksum"):
                with CubexParser(cubex_file_path) as cube_file:
                    cube_file.all_metrics()
        except FileNotFoundError as err:
            raise SkipTest("Required test file was not found.") from err


if __name__ == '__main__':
    unittest.main()



# This file is part of the Extra-P software (http://www.scalasca.org/software/extra-p)
#
# Copyright (c) 2020-2022, Technical University of Darmstadt, Germany
#
# This software may be modified and distributed under the terms of a BSD-style license.
# See the LICENSE file in the base directory for details.

import unittest
from pathlib import Path
from unittest import SkipTest

from pycubexr import CubexParser


class TestExtendedCudaKernels(unittest.TestCase):
    cubex: CubexParser = None

    @classmethod
    def setUpClass(cls) -> None:
        try:
            cubex_file_path = Path("../data/cube_samples/profile_extended_kernel.cubex").resolve(strict=True)
        except FileNotFoundError as err:
            raise SkipTest("Required test file was not found.") from err
        cls.cubex = CubexParser(cubex_file_path).__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def test_method_names_not_mangled(self):
        cnodes = self.cubex.all_cnodes()
        print(cnodes[15].region.name)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pycubexr.classes import Metric


class MissingMetricError(Exception):
    def __init__(self, metric: Metric):
        self.metric = metric
        super().__init__('The cubex file does NOT contain an index and/or data for the metric ({})'.format(metric))


class UnsupportedMetricFormatError(Exception):
    def __init__(self, metric_format: str, metric: Metric = None):
        self.metric_format = metric_format
        self.metric = metric
        message = 'The cubex parser does NOT support the format ({})'.format(metric_format)
        if metric is not None:
            message += ' of this metric: {}'.format(metric)
        super().__init__(message)


class CorruptIndexError(Exception):
    def __init__(self, message='Corrupt index.'):
        super().__init__(message)


class InvalidConversionInstructionError(Exception):
    def __init__(self, message="Received invalid conversion instruction. "
                               "convert_to_exclusive and convert_to_inclusive must not be True at the same time."):
        super().__init__(message)



from typing import List, Any

from pycubexr.utils.logger import logger


def chunk_list(elements: List[Any], chunk_size: int):
    return [elements[x: x + chunk_size] for x in range(0, len(elements), chunk_size)]



import copy
import sys
from abc import abstractmethod, ABC
from numbers import Real, Number

from typing import List, Dict, Callable, Union, Sequence, Any, cast, Iterable

import numpy as np


class CubeValues:

    def __init__(self, values):
        self._values: np.ArrayLike = values

    def __add__(self, other):
        if isinstance(other, type(self)):
            return type(self)(self._values + other._values)
        else:
            return NotImplemented

    def __sub__(self, other):
        if isinstance(other, type(self)):
            return type(self)(self._values - other._values)
        else:
            return NotImplemented

    def __repr__(self):
        return f"{type(self).__name__}({self._values})"

    def __len__(self):
        return len(self._values)

    def __getitem__(self, item):
        return type(self)(self._values[item])

    def sum(self):
        return type(self)(np.sum(self._values))

    def mean(self):
        return type(self)(np.mean(self._values))

    def astype(self, type):
        return self._values.astype(type)

    def copy(self):
        return type(self)(self._values.copy())


class MinMaxValues(CubeValues, ABC):

    @classmethod
    @abstractmethod
    def agg_command(cls, a, b, out=None):
        pass

    def _check(self, other):
        if all(other._values == 0):
            return self
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return None

    def __add__(self, other):
        other_values = self._check(other)
        if other_values is not None:
            return other_values
        return type(self)(self.agg_command(self._values, other._values))

    def __sub__(self, other):
        other_values = self._check(other)
        if other_values is not None:
            return other_values
        return type(self)(self.agg_command(self._values, other._values))

    def __iadd__(self, other):
        other_values = self._check(other)
        if other_values is not None:
            return other_values
        self.agg_command(self._values, other._values, out=self._values)
        return self

    def __isub__(self, other):
        other_values = self._check(other)
        if other_values is not None:
            return other_values
        self.agg_command(self._values, other._values, out=self._values)
        return self


class MinValues(MinMaxValues):
    agg_command = np.minimum

    def sum(self):
        return MinValues(self._values.min())

    def mean(self):
        return MinValues(self._values.min())


class MaxValues(MinMaxValues):
    agg_command = np.maximum

    def sum(self):
        return MaxValues(self._values.max())

    def mean(self):
        return MaxValues(self._values.max())


class ComplexValues(CubeValues):

    def astype(self, type):
        if type != complex:
            return np.absolute(self._values).astype(type)
        return self._values.astype(type)


class TauAtomicValues(CubeValues):
    def astype(self, type):
        raise NotImplementedError()


class RateValues(CubeValues):

    def astype(self, type):
        raise NotImplementedError()


# class HistogramValue(BaseValue):
#     def __init__(self, value):
#         if isinstance(value, HistogramValue):
#             self.min, self.max = value.min, value.max
#             self.values = copy.copy(value.values)
#         else:
#             self.min, self.max = value[0:2]
#             self.values = value[2:]
#
#     @property
#     def n(self):
#         return len(self.values)
#
#     def sub(self, other):
#         return self
#
#     def neutral(self):
#         return HistogramValue(
#             [-math.inf, math.inf]
#         )
#
#     def add(self, other):
#         return HistogramValue(
#             [min(self.min, other.min), max(self.max, other.max)] +
#             self.values + other.values
#         )


VALUE_MAPPING: Dict[str, Callable[[Union[Sequence[Any], Any]], Any]] = {
    'MAXDOUBLE': MaxValues,
    'MINDOUBLE': MinValues,
    'COMPLEX': ComplexValues,
    'TAU_ATOMIC': TauAtomicValues,
    'RATE': RateValues,
    # 'NDOUBLES': NValues,
    # 'HISTOGRAM': HistogramValue
}

# def _determine_max_uint64_for_double():
#     for i in range(0xFFFF_FFFF_FFFF_FFFF, 0, -1):
#         if int(float(i)) < int(float(0xFFFF_FFFF_FFFF_FFFF)):
#             # print(f"{i:x}")
#             return i

MAX_UINT64_DOUBLE = 0xFFFF_FFFF_FFFF_FBFF  # Maximal UINT64 that can be cast to double and back without exceeding UINT64


def convert_type(type_: str, parameters: tuple, values: Iterable[Union[tuple, Real]],
                 allow_full_uint64_values: bool = False) -> List[Union[CubeValues, Real]]:
    if type_ in VALUE_MAPPING:
        val_cls = VALUE_MAPPING[type_]
        # if 'type_params' in signature(val_cls).parameters:
        #     # noinspection PyArgumentList
        #     return [val_cls(v, type_params=parameters) for v in values]
        # else:
        return val_cls(values)
    elif isinstance(values, np.ndarray):
        if not allow_full_uint64_values and type_ == 'UINT64' or type_ == 'UNSIGNED INTEGER':
            if any(values > MAX_UINT64_DOUBLE):
                # simulates undefined behavior when casting uint64_t to double to uint64_t in cubelib
                values = np.require(values, requirements='W')
                values[values > MAX_UINT64_DOUBLE] = 0
        return values
    elif not allow_full_uint64_values and type_ == 'UINT64' or type_ == 'UNSIGNED INTEGER':
        # simulates undefined behavior when casting uint64_t to double to uint64_t in cubelib
        return [0 if value > MAX_UINT64_DOUBLE else value for value in values]
    else:

        return list(values)



import unittest
from pathlib import Path
from typing import Callable

import numpy as np

from pycubexr import CubexParser
from pycubexr.utils.exceptions import UnsupportedMetricFormatError
from pycubexr.utils.metric_formats import METRIC_FORMATS


class TestMetricTypes(unittest.TestCase):
    def test_list_of_metric_types(self):
        for metric_type in METRIC_FORMATS.values():
            if isinstance(metric_type, Callable):
                metric_type = metric_type('5')

            print(metric_type)
            test = np.dtype(metric_type)
            print(metric_type, ':', test)

    def test_metric_types(self):
        cubex_file_path = Path("../data/test_only_files/cube-values_example.cubex").resolve()
        with CubexParser(cubex_file_path) as cubex:
            for metric in cubex.get_metrics():
                with self.subTest(metric=metric.name):
                    if metric.name == "HISTOGRAM5":
                        self.assertEqual("HISTOGRAM(5)", metric.data_type)
                    elif metric.name == "NDOUBLES10":
                        self.assertEqual("NDOUBLES(10)", metric.data_type)
                    elif metric.name == "SCALE_FUNC3":
                        self.assertEqual("SCALE_FUNC", metric.data_type)
                    elif metric.name == "TAU_ATOMIC2":
                        self.assertEqual("TAU_ATOMIC", metric.data_type)
                    else:
                        self.assertEqual(metric.name, metric.data_type)

    def test_parsing(self):
        cubex_file_path = Path("../data/test_only_files/cube-values_example.cubex").resolve()
        with CubexParser(cubex_file_path) as cubex:
            for metric in cubex.get_metrics():
                with self.subTest(metric=metric.name):
                    if metric.name in ["HISTOGRAM5", "SCALE_FUNC3"]:
                        self.assertRaises(UnsupportedMetricFormatError, cubex.get_metric_values, metric)
                    else:
                        cubex.get_metric_values(metric)


if __name__ == '__main__':
    unittest.main()



import warnings
from typing import List


class MetricType(object):
    INCLUSIVE = 'INCLUSIVE'
    EXCLUSIVE = 'EXCLUSIVE'


class Metric(object):

    def __init__(
            self,
            *,
            name: str,
            _id: int,
            display_name: str,
            description: str,
            metric_type: str,
            data_type: str,
            units: str,
            url: str,
            childs: List
    ):
        assert hasattr(MetricType, metric_type)
        self.name = name
        self.id = _id
        self.display_name = display_name
        self.description = description
        self.metric_type = metric_type
        self.data_type = data_type
        self.units = units
        self.url = url
        self.tree_index_to_cid_map: dict = None
        self.childs = childs

    @property
    def tree_enumeration(self):
        warnings.warn('Accessing the tree enumeration is deprecated. Use cnode ids directly.', DeprecationWarning)
        if self.tree_index_to_cid_map is None:
            return None
        return {cid: tid for tid, cid in self.tree_index_to_cid_map.items()}

    def __repr__(self):
        return 'Metric<{}>'.format(self.__dict__)

    def get_all_children(self, with_self=True) -> List['Metric']:
        metrics = []
        if with_self:
            metrics.append(self)
        for child in self.childs:
            metrics += child.get_all_children()
        return metrics



import sys

if sys.version_info >= (3, 8):
    import functools

    cached_property = functools.cached_property
else:
    class cached_property(object):
        """
        A property that is cached for later retrieval.
        """

        def __init__(self, func):
            self.func = func

        def __get__(self, obj, cls):
            if obj is None:
                return self

            value = self.func(obj)
            obj.__dict__[self.func.__name__] = value
            return value



import unittest

import numpy as np
from pycubexr.classes.metric_values import MetricValues


class TestOverflowDetection(unittest.TestCase):

    def test_subtract_int_16_8(self):
        type_ = np.int16
        a = np.arange(np.iinfo(type_).min, np.iinfo(type_).max, dtype=type_)
        b = np.full_like(a, 1, dtype=np.int8)
        b = MetricValues._detect_negative_overflow(a, b)
        self.assertSequenceEqual([1] * (np.iinfo(type_).max - np.iinfo(type_).min), b.tolist())
        res = a - b
        self.assertSequenceEqual(list(range(np.iinfo(type_).min - 1, np.iinfo(type_).max - 1)), res.tolist())

    def test_subtract_int_64_8(self):
        type_ = np.int64
        a = np.arange(np.iinfo(type_).min, np.iinfo(type_).min + 10000, dtype=type_)
        b = np.full_like(a, 1, dtype=np.int8)
        b = MetricValues._detect_negative_overflow(a, b)
        self.assertSequenceEqual([0] + [1] * (10000 - 1), b.tolist())
        res = a - b
        self.assertSequenceEqual(
            [np.iinfo(type_).min] + list(range(np.iinfo(type_).min, np.iinfo(type_).min + 10000 - 1)), res.tolist())

    def test_subtract_int_64_16(self):
        type_ = np.int64
        a = np.arange(np.iinfo(type_).min, np.iinfo(type_).min + 1000, dtype=type_)
        b = np.full_like(a, 100, dtype=np.int16)
        b = MetricValues._detect_negative_overflow(a, b)
        self.assertSequenceEqual(list(range(0, 100)) + [100] * (1000 - 100), b.tolist())
        res = a - b
        self.assertSequenceEqual(
            [np.iinfo(type_).min] * 100 + list(range(np.iinfo(type_).min, np.iinfo(type_).min + 1000 - 100)),
            res.tolist())

    def test_subtract_zero_int_64_16(self):
        type_ = np.int64
        a = np.arange(np.iinfo(type_).min, np.iinfo(type_).min + 1000, dtype=type_)
        b = np.zeros_like(a, dtype=np.int16)
        b = MetricValues._detect_negative_overflow(a, b)
        self.assertSequenceEqual([0] * (1000), b.tolist())
        res = a - b
        self.assertSequenceEqual(a.tolist(), res.tolist())

    def test_subtract_negative_int_64_16(self):
        type_ = np.int64
        a = np.arange(np.iinfo(type_).max - 1000, np.iinfo(type_).max + 1, dtype=type_)
        b = np.full_like(a, -100, dtype=np.int16)
        b = MetricValues._detect_negative_overflow(a, b)
        self.assertSequenceEqual([-100] * (1000 - 100) + list(range(-100, 1)), b.tolist())
        res = a - b
        self.assertSequenceEqual(
            list(range(np.iinfo(type_).max - 1000 + 100, np.iinfo(type_).max)) + [np.iinfo(type_).max] * 101,
            res.tolist())



from xml.etree.ElementTree import Element as XMLNode, ElementTree

from pycubexr.classes import CNode, Location, LocationGroup, Metric, Region, SystemTreeNode


def parse_metric(xml_node: XMLNode):
    return Metric(
        name=xml_node.findtext('uniq_name'),
        _id=int(xml_node.get('id')),
        display_name=xml_node.findtext('disp_name'),
        description=xml_node.findtext('descr'),
        metric_type=xml_node.get('type'),
        data_type=xml_node.findtext('dtype'),
        units=xml_node.findtext('uom'),
        url=xml_node.findtext('url'),
        childs=parse_metrics(xml_node)
    )


def parse_metrics(root: XMLNode):
    if isinstance(root, ElementTree):
        root = root.find('metrics')
    return [
        parse_metric(metric_xml_node) for metric_xml_node
        in root.findall('metric')
    ]


def parse_region(xml_node: XMLNode):
    return Region(
        _id=int(xml_node.get('id')),
        begin=int(xml_node.get('begin')),
        end=int(xml_node.get('end')),
        name=xml_node.findtext('name'),
        mangled_name=xml_node.findtext('mangled_name'),
        paradigm=xml_node.findtext('paradigm'),
        role=xml_node.findtext('role'),
        url=xml_node.findtext('url'),
        descr=xml_node.findtext('descr'),
        mod=xml_node.get("mod"),
    )


def parse_regions(root: XMLNode):
    return [
        parse_region(xml_node) for xml_node
        in root.find('program').findall('region')
    ]


def parse_attrs(root: XMLNode):
    return {
        node.get('key'): node.get('value') for node
        in root.findall('attr')
    }


def parse_cnode(xml_node: XMLNode):
    cnode = CNode(
        _id=int(xml_node.get('id')),
        callee_region_id=int(xml_node.get('calleeId'))
    )
    for cnode_xml_child in xml_node.findall('cnode'):
        cnode_child = parse_cnode(cnode_xml_child)
        cnode_child.parent = cnode
        cnode.add_child(cnode_child)
    for parameter_xml in xml_node.findall('parameter'):
        partype = parameter_xml.get('partype')
        parkey = parameter_xml.get('parkey')
        parvalue = parameter_xml.get('parvalue')
        if partype == 'numeric':
            try:
                parvalue = int(parvalue)
            except ValueError:
                parvalue = float(parvalue)
        elif partype == "string":
            pass
        cnode.parameters[parkey] = parvalue
    return cnode


def parse_cnodes(root: XMLNode):
    return [
        parse_cnode(cnode) for cnode
        in root.find('program').findall('cnode')
    ]


def parse_location(xml_node: XMLNode):
    return Location(
        _id=int(xml_node.get('Id')),
        name=xml_node.findtext('name'),
        rank=xml_node.findtext('rank'),
        _type=xml_node.findtext('type')
    )


def parse_location_group(xml_node: XMLNode):
    location_group = LocationGroup(
        _id=int(xml_node.get('Id')),
        name=xml_node.findtext('name'),
        rank=xml_node.findtext('rank'),
        _type=xml_node.findtext('type')
    )

    for xml_child_node in xml_node.findall('locationgroup'):
        location_group.add_location_group(parse_location_group(xml_child_node))

    for xml_child_node in xml_node.findall('location'):
        location_group.add_location(parse_location(xml_child_node))

    return location_group


def parse_system_tree_node(xml_node: XMLNode):
    system_tree_node = SystemTreeNode(
        _id=int(xml_node.get('Id')),
        _class=xml_node.get('class'),
        name=xml_node.findtext('name'),
        attrs={x.get('key'): x.get('value') for x in xml_node.findall('attr')}
    )

    for xml_child_node in xml_node.findall('systemtreenode'):
        system_tree_node.add_system_tree_node_child(parse_system_tree_node(xml_child_node))

    for xml_child_node in xml_node.findall('locationgroup'):
        system_tree_node.add_location_group(parse_location_group(xml_child_node))

    return system_tree_node


def parse_system_tree_nodes(root: XMLNode):
    return [
        parse_system_tree_node(xml_node) for xml_node
        in root.find('system').findall('systemtreenode')
    ]



import unittest
from pathlib import Path

from pycubexr import CubexParser
from pycubexr.classes.metric import MetricType
from pycubexr.utils.exceptions import MissingMetricError


class TestBasicCallTree(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cubex_file_path = Path("../data/call_tree_test/call_tree_test.cubex").resolve()
        cls.cubex = CubexParser(cubex_file_path).__enter__()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.cubex.__exit__(None, None, None)

    def test_root_cnode(self):
        root_cnode = self.cubex.get_root_cnodes()
        self.assertEqual(len(root_cnode), 1)
        self.assertEqual(0, root_cnode[0].id)
        self.assertEqual('test.x', root_cnode[0].region.name)
        self.assertEqual(2, root_cnode[0].region.id)

    def test_callpaths(self):
        callpaths = self._walk_cnode_tree(self.cubex.get_root_cnodes())
        expected_callpaths = {i: v for i, v in enumerate([
            "test.x",
            "test.x->main",
            "test.x->main->a",
            "test.x->main->a->a1",
            "test.x->main->a->a2",
            "test.x->main->a->a3",
            "test.x->main->b",
            "test.x->main->b->b1",
            "test.x->main->b->b2",
            "test.x->main->b->b3",
            "test.x->main->c",
            "test.x->main->c->c1",
            "test.x->main->c->c2",
            "test.x->main->c->c3",
            "test.x->main->d",
            "test.x->main->d->d1",
            "test.x->main->d->d2",
            "test.x->main->d->d3",
        ])}
        self.assertDictEqual(expected_callpaths, callpaths)

    def test_cnodes(self):
        expected_cnodes = [
            "cube_test_file/test.x",
            "main",
            "a",
            "a1",
            "a2",
            "a3",
            "b",
            "b1",
            "b2",
            "b3",
            "c",
            "c1",
            "c2",
            "c3",
            "d",
            "d1",
            "d2",
            "d3",
        ]
        for i, v in enumerate(expected_cnodes):
            cnode = self.cubex.get_cnode(i)
            self.assertEqual(v, cnode.region.mangled_name)

    def _walk_cnode_tree(self, cnodes):

        callpaths = {}

        def walk_tree(parent_cnode, parent_name, level):
            num_nodes_per_level = {0: 1, 1: 4, 2: 3, 3: 0}
            self.assertEqual(len(parent_cnode.get_children()), num_nodes_per_level[level])

            for cnode in parent_cnode.get_children():
                name = cnode.region.mangled_name
                path_name = '->'.join((parent_name, name))
                callpath = path_name
                callpaths[cnode.id] = callpath
                walk_tree(cnode, path_name, level + 1)

        for root_cnode in cnodes:
            name = root_cnode.region.name
            callpath = name
            callpaths[root_cnode.id] = callpath
            walk_tree(root_cnode, name, 0)

        return callpaths

    def test_metrics(self):
        metrics = ['visits', 'time', 'min_time', 'max_time', 'bytes_put', 'bytes_get', 'io_bytes_read',
                   'io_bytes_written']
        metric_types = [MetricType.EXCLUSIVE, MetricType.INCLUSIVE, MetricType.EXCLUSIVE, MetricType.EXCLUSIVE,
                        MetricType.EXCLUSIVE, MetricType.EXCLUSIVE, MetricType.EXCLUSIVE, MetricType.EXCLUSIVE]
        for i, cube_metric in enumerate(self.cubex.get_metrics()):
            self.assertEqual(cube_metric.name, metrics[i])
            self.assertEqual(i, cube_metric.id)
            self.assertEqual(cube_metric.metric_type, metric_types[i])

    def test_values_visits(self):
        metric = self.cubex.get_metric_by_name('visits')
        metric_values = self.cubex.get_metric_values(metric, cache=False)
        expected_values = [
            1,
            1,
            1,
            1, 2, 3,
            2,
            2, 4, 6,
            3,
            3, 6, 9,
            4,
            4, 8, 12,
        ]
        for i, v in enumerate(expected_values):
            cnode = self.cubex.get_cnode(i)
            values = metric_values.cnode_values(cnode)
            self.assertEqual(v, values[0])

    def test_values_time(self):
        metric = self.cubex.get_metric_by_name('time')
        metric_values = self.cubex.get_metric_values(metric, cache=False)
        expected_values = [
            74.05,
            74.05,
            60.00,
            10.0, 20.0, 30.0,
            12.00,
            2.0, 4.0, 6.0,
            1.80,
            0.3, 0.6, 0.9,
            0.24,
            0.04, 0.08, 0.12,
        ]
        for i, v in enumerate(expected_values):
            cnode = self.cubex.get_cnode(i)
            values = metric_values.cnode_values(cnode)
            self.assertAlmostEqual(v, values[0], 2)

    def test_values_time2(self):
        metric = self.cubex.get_metric_by_name('time')
        metric_values = self.cubex.get_metric_values(metric, cache=False)
        expected_values = [
            74.05,
            74.05,
            60.00,
            10.0, 20.0, 30.0,
            12.00,
            2.0, 4.0, 6.0,
            1.80,
            0.3, 0.6, 0.9,
            0.24,
            0.04, 0.08, 0.12,
        ]

        print(metric_values.values)

        for i, v in enumerate(expected_values):
            cnode = self.cubex.get_cnode(i)
            values = metric_values.cnode_values(cnode)
            self.assertAlmostEqual(v, values[0], 2)

    def test_values_time_exclusive(self):
        metric = self.cubex.get_metric_by_name('time')
        metric_values = self.cubex.get_metric_values(metric, cache=False)
        expected_values = [
            74.05,
            74.05,
            60.00,
            10.0, 20.0, 30.0,
            12.00,
            2.0, 4.0, 6.0,
            1.80,
            0.3, 0.6, 0.9,
            0.24,
            0.04, 0.08, 0.12,
        ]
        for i, v in enumerate(expected_values):
            cnode = self.cubex.get_cnode(i)
            values = metric_values.cnode_values(cnode)
            self.assertAlmostEqual(v, values[0], 2)


class TestCallTree(unittest.TestCase):

    def test_bytes_recieved_problematic_case(self):
        cubex_file_path = Path("../data/time.p4.n2000.x1.r0/profile.cubex").resolve()
        with CubexParser(cubex_file_path) as parsed:
            root_cnodes = parsed.get_root_cnodes()
            callpaths = self._make_callpath_mapping(root_cnodes)
            cube_metric = parsed.get_metric_by_name('bytes_received')
            metric_values = parsed.get_metric_values(metric=cube_metric, cache=False)
            last_mpi_recv = list(metric_values.cnode_indices.keys())[-1]
            # self.assertEqual(last_mpi_recv, 45)
            cnode = parsed.get_cnode(45)
            print(cnode)

    def test_bytes_sent(self):
        cubex_file_path = Path("../data/time.p4.n2000.x1.r0/profile.cubex").resolve()
        with CubexParser(cubex_file_path) as parsed:
            root_cnodes = parsed.get_root_cnodes()
            callpaths, _ = self._make_callpath_mapping(root_cnodes)
            cube_metric = parsed.get_metric_by_name('bytes_sent')
            try:
                metric_values = parsed.get_metric_values(metric=cube_metric, cache=False)
                expected_cnode_ids = [10, 13, 16, 19, 22, 25, 28, 31, 34, 38]
                self.assertListEqual(expected_cnode_ids, list(metric_values.cnode_indices.keys()))
                for cnode_id in metric_values.cnode_indices:
                    cnode = parsed.get_cnode(cnode_id)
                    # self.assertEqual(sorted_ids[cnode_id], cnode_id)
                    callpath = callpaths[cnode_id]
                    self.assertTrue(callpath.endswith(cnode.region.name))
            except MissingMetricError:
                print("Missing metric for", cube_metric.name)

    def test_bytes_recieved(self):
        cubex_file_path = Path("../data/time.p4.n2000.x1.r0/profile.cubex").resolve()
        with CubexParser(cubex_file_path) as parsed:
            root_cnodes = parsed.get_root_cnodes()
            callpaths, _ = self._make_callpath_mapping(root_cnodes)
            cube_metric = parsed.get_metric_by_name('bytes_received')
            try:
                metric_values = parsed.get_metric_values(metric=cube_metric, cache=False)
                expected_cnode_ids = [10, 13, 16, 19, 22, 25, 28, 31, 44, 45]
                self.assertListEqual(expected_cnode_ids, list(metric_values.cnode_indices.keys()))
                for cnode_id in metric_values.cnode_indices:
                    cnode = parsed.get_cnode(cnode_id)
                    # self.assertEqual(sorted_ids[cnode_id], cnode_id)
                    callpath = callpaths[cnode_id]
                    self.assertTrue(callpath.endswith(cnode.region.name))
            except MissingMetricError:
                print("Missing metric for", cube_metric.name)

    def _make_callpath_mapping(self, cnodes):

        sort_callpaths_id = [0]  # Ensures that the callpath ID is compatible for the next steps
        callpaths = {}

        def walk_tree(parent_cnode, parent_name):
            for cnode in parent_cnode.get_children():
                name = cnode.region.name
                path_name = '->'.join((parent_name, name))
                callpath = path_name
                callpaths[cnode.id] = callpath
                sort_callpaths_id.append(cnode.id)
                walk_tree(cnode, path_name)

        for root_cnode in cnodes:
            name = root_cnode.region.name
            callpath = name
            callpaths[root_cnode.id] = callpath
            walk_tree(root_cnode, name)

        return callpaths, sort_callpaths_id


if __name__ == '__main__':
    unittest.main()
